package stamp.math;
import stamp.util.text.*;

/**
 * This class implements CORDIC (COordinate Rotation DIgital Computer) algorithmes to calculate
 * sin(x), cos(x), tan(x), len(y,x) and atan2(y,x).
 * The methods are adapted from the cordic programs found at http://www.emesystems.com/BS2mathC.htm
 * Cordic algorithmes do not require multiplication and division, only adds, subtracts and shifts.
 *
 * @author Peter Verkaik (peterverkaik@boselectro.nl)
 * @version 1.0 December 25, 2005
 */

/*
  In the most general form one CORDIC iteration can be written as

    x[i+1] = x[i] - m*d[i]*y[i]*(2^-i)
    y[i+1] = y[i] + d[i]*x[i]*(2^-i)
    z[i+1] = z[i] - d[i]*a[i]

  It may not be immediately obvious this basic CORDIC iteration describes a rotation (together with a scaling)
  of an intermediate plane vector v[i]=(x[i],y[i]) to v[i+1]=(x[i+1],y[i+1]).
  The third iteration variable z[i] keeps track of the rotation angle a[i].
  The variable m={1,0,-1} specifes a circular, linear or hyperbolic coordinate system, respectively.
  The rotation direction is steered by the variable d={1,-1}.

  sine and cosine
    x[0] = unit vector length          x[n] = cos(x)
    y[0] = 0                           y[n] = sin(x)
    z[0] = angle                       z[n] = 0

  polar to rectangular transformation
    x[0] = vector magnitude            x[n] = x-coordinate of x,y vector
    y[0] = 0                           y[n] = y-coordinate of x,y vector
    z[0] = polar angle                 z[n] = 0

  arctangent and vector length
    x[0] = x-coordinate of x,y vector  x[n] = vector length (polar magnitude)
    y[0] = y-coordinate of x,y vector  y[n] = 0
    z[0] = 0                           z[n] = arctangent (polar angle)

  lineair functions
    x[0] = x                           x[n] = x[0]
    y[0] = y                           y[n] = 0
    z[0] = 0                           z[n] = -(y[0]/x[0])
*/

public class cordic {

  private static int xreg; //x accumulator
  private static int yreg; //y accumulator

  //cordic gain K = 1/1.64676 for n->infinityreg 1/1.64676 = 0.607253 => .607253*645536 = 39797
  private static final int GAIN = (short)39798;
  private static int[] atnTable = new int[]{16384,9672,5110,2594,1302,652,326,163,81,41,20,10,5,3,1,0};
                                  //16384/45*atn(1/1),16384/45*atn(1/2),16384/45*atn(1/4)...16384/45*atn(1/32768)
  /**
   * Calculate sin(x)
   *
   * @param x Angle in 0.1 degree units (-3600 to +3600)
   * @return sin(x) in 0.0001 units (-10000 to +10000)
   */
  public static int sin(int x) {
    sincos(x);
    return yreg;
  }

  /**
   * Calculate cos(x)
   *
   * @param x Angle in 0.1 degree units (-3600 to +3600)
   * @return cos(x) in 0.0001 units (-10000 to +10000)
   */
  public static int cos(int x) {
    sincos(x);
    return xreg;
  }

  /*
     We want tan(x) in 0.01 units
     100*tan(x) = 100*sin(x)/cos(x) = 100*yreg/xreg = (100/xreg)*yreg
     100/xreg = I + F/65536
     100*tan(x) = yreg*I + UnsignedIntMath.umulf(yreg,F)
  */

  /**
   * Calculate tan(x)
   *
   * @param x Angle in 0.1 degree units (-3600 to +3600)
   * @return tan(x) in 0.01 units (-29409 to +29409 for angles -89.8 to +89.8)
   *         angles 89.9,90.0 and -90.1 return 32767, angles -89.9,-90.0 and 90.1 return -32767
   */
  public static int tan(int x) {
    x = (x + 3600)%3600;
    int q = x%900;
    switch (x/900) {
      case 1: return (q == 0) ? 32767 : -tan(900-q);
      case 2: return tan(q);
      case 3: return (q == 0) ? -32767 : -tan(900-q);
    }
    if (q == 899) return 32767;
    int a = sin(900-q); //cos(q)
    int b = sin(q);
    //tan(x)=100*sin(x)/cos(x)=(I + F/65536)*sin(x)
    return ((100/a)*b) + UnsignedIntMath.umulf(b,UnsignedIntMath.ufrac(100,a));
  }

  /**
   * Calculate length of vector y,x
   * Absolute max value of x and y is 23170 because sqrt(23170^2 + 23170^2) = 32767
   *
   * @param y Vertical signed value (-10000 to +10000)
   * @param x Horizontal signed value (-10000 to +10000)
   * @return length of vector
   */
  public static int len(int y, int x) {
    atnlen(y,x);
    return xreg;
  }

  /**
   * Calculate arctan(y/x)
   * Absolute max value of x and y is 23170 because sqrt(23170^2 + 23170^2) = 32767
   *
   * @param y Vertical signed value (-10000 to +10000)
   * @param x Horizontal signed value (-10000 to +10000)
   * @return Angle in 0.1 degree units (-1799 to +1800)
   */
  public static int atan2(int y, int x) {
    atnlen(y,x);
    return yreg;
  }

  /**
   * Calculate value shifted right n bits with sign preserved
   *
   * @param reg Value to be shifted right
   * @param n Number of bits to shift right
   * @return Shifted right value with sign preserved
   */
  private static int delta(int reg, int n) {
    int s = reg>>>15;
    return  ((-s) ^ (UnsignedIntMath.abs(reg) >> n)) + s;
  }

  /**
   * Calculate arctangent and vectorlength using cordic vector mode
   * This method returns len(y,x) in regx, atan2(y,x) in regy
   *
   * @param y Vertical signed value
   * @param x Horizontal signed value
   */
  private static void atnlen(int y, int x) {
    int i,atn,xs,ys,zs,xx,z,q23,prescale=0;
    //check for x=0
    if (x == 0) {
      xreg = UnsignedIntMath.abs(y); //vector length equals abs(y)
      if (y == 0) yreg = 450; //treat as y==x
      else yreg = (y < 0) ? -900 : 900;
      return;
    }
    xreg = x;
    yreg = y;
    q23 = xreg>>>15;
    if (q23 == 1) xreg = -xreg; //move to 1st or 4th quadrant, where cordic works
    xs = UnsignedIntMath.abs(xreg);
    ys = UnsignedIntMath.abs(yreg);
    //prescale for large input values
    if (xs>19898 || ys>19898) { //19898 = 32767/1.64676
      xs >>>= 1;
      ys >>>= 1;
      prescale = -1;
      xreg >>= 1; //xreg /= 2; remember xreg > 0 at this point
      yreg = delta(yreg,1); //yreg /= 2; divide by 2 with sign preserved
    }
    //prescale for small input values
    if (xs<16384 && ys<16384) { //allow for at least one shift
      while ((xs<<1)<10001 && (ys<<1)<10001 && prescale<7) {
        xs <<= 1;
        ys <<= 1;
        prescale++;
        xreg <<= 1;  //xreg *= 2; remember xreg > 0 at this point
        i = yreg>>>15;
        yreg = ((-i) ^ (UnsignedIntMath.abs(yreg) << 1)) + i; //yreg *= 2; multiply by 2 with sign preserved
      }
    }
    z = 0;
    for (i=0; i<16; i++) {
      atn = atnTable[i];
      xx = xreg + (UnsignedIntMath.abs(yreg) >> i);
      if (yreg<0) {
        //yreg<0, ys=1, d=+1
        yreg = yreg + delta(xreg,i); //yreg<0 on entry
        z = z - atn;
      }
      else {
        //yreg>=0, ys=0, d=-1
        yreg = yreg - delta(xreg,i); //yreg>=0 on entry
        z = z + atn;
      }
      xreg = xx;
    }
    zs = z>>>15;
    z = ((-zs) ^ UnsignedIntMath.umulf(UnsignedIntMath.abs(z),1800)) + zs;
    if (q23 == 1) { //wrapper for 2nd and 3rd quadrants
      if (z<0) z = -1800 - z;
      else z = 1800 - z;
    }
    yreg = z; //put angle -179.9 to +180.0 result in yreg
    xreg = UnsignedIntMath.umulf(xreg,GAIN); //vector has been rotated to x axis, multiply by CORDIC gain 0.607
    //undo prescale (only affects xreg)
    while (prescale > 0) {
      xreg = (xreg+1)>>>1;
      prescale--;
    }
    while (prescale < 0) {
      xreg <<= 1;
      prescale++;
    }
  }

  /**
   * Calculate sine and cosine using cordic rotation mode.
   * This method returns cos(x) in xreg, sin(x) in yreg
   * Note: no multiplication or division is used in this method!
   *
   * @param x Angle in 0.1 degree units (-3600 to +3600)
   */
  private static void sincos(int x) {
    int i,atn,xs,ys,zs,xx,z,q;
    z = (x+3600)%3600;      //translate angle to 0.0 to 360.0 degrees
    q = z / 900;            //what quadrant?  save as 0,1,2,3
    switch (q) {
      case 1:
      case 2: z = z - 1800; //rotate 90.0 to 270.0 degrees to -90.0 to +90.0 degrees
              break;
      case 3: z = z - 3600; //4th quadrant will be 0.0 to -90.0 degrees.
    }
    zs = z>>>15; //translate angle -90.0 to +90.0 to -32768 to +32767
    z = ((-zs)^(UnsignedIntMath.abs(z)*36 + UnsignedIntMath.umulf(UnsignedIntMath.abs(z),26797))) + zs;
                                     //36 = int(16384/450)   26797 = 65536*frac(16384/450)
    yreg = 0;      //initial unit vector length one (=10000)
    xreg = 10000;  //along x axis
    for (i=0; i<16; i++) {
      atn = atnTable[i];
      if (z<0) {
        //yreg<0, ys=1, d=+1
        xx = xreg + delta(yreg,i); //yreg<0 on entry
        yreg = yreg - delta(xreg,i);
        z = z + atn;
      }
      else {
        //yreg>=0, ys=0, d=-1
        xx = xreg - delta(yreg,i); //yreg>=0 on entry
        yreg = yreg + delta(xreg,i);
        z = z - atn;
      }
      xreg = xx;
    }
    xs = xreg>>>15;
    xreg = ((-xs) ^ UnsignedIntMath.umulf(UnsignedIntMath.abs(xreg),GAIN)) + xs; //multiply to compensate CORDIC gain 0.607
    ys = yreg>>>15;
    yreg = ((-ys) ^ UnsignedIntMath.umulf(UnsignedIntMath.abs(yreg),GAIN)) + ys; //multiply to compensate CORDIC gain 0.607
    if (q==1 || q==2) {
      xreg = -xreg;
      yreg = -yreg;
    }
  }

}

