package stamp.math;
import stamp.core.*;
import stamp.util.text.*;

/**
 * This class provides floating point math routines, limited to 3 digits accuracy.
 * A floating point value is stored in an integer (16 bits), to keep the code minimal.
 * This should fill the gap between fixed point arithmatic (very limited range) and
 * the true floating point class Float32 (huge code).
 *
 * f3 value floating point format
 *
 *     +-+----------+-+----+
 *     |s|mmmmmmmmmm|x|eeee|
 *     +-+----------+-+----+
 *
 *     s = mantissa sign (1 bit)
 *     m = mantissa      (10 bits)
 *     x = exponent sign (1 bit)
 *     e = exponent      (4 bits)
 *
 * The signed range is from -0.999E+15 to +0.999E+15 which should be sufficient.
 * Near zero, the smallest values are -0.001E-16 and +0.001E-16 which also should be sufficient.
 * The f3 value zero 0.000E+00 equals the integer value zero (binary null).
 *
 * Given the fact, that if a measured value has a 3 digit accuracy (like ADC values),
 * any result calculated with that value can never have more significant digits, the 3 digit
 * accuracy of a f3 value may be accurate enough for those cases. If it isn't, use the
 * Float32 package or a co-processor like the uM-FPU.
 *
 * @author Peter Verkaik (peterverkaik@boselectro.nl)
 * @version 1.0 January 4, 2006
 */

public class f3 {

  /**
   * An identifier for the positive infinity constant.
   */
  public static final int POSITIVE_INFINITY = (999<<5)|15; //+0.999E+15

  /**
   * An identifier for the negative infinity constant.
   */
  public static final int NEGATIVE_INFINITY = (-999<<5)|15; //-0.999E+15

  /**
   * An identifier for the Not-a-Number constant (NaN).
   */
  public static final int NaN = -1;

  /**
   * An identifier for the constant Pi.
   */
  public static final int PI = (314<<5)|1; //3.14

  /**
   * An identifier for the constant e.
   */
  public static final int E = (272<<5)|1; //2.72

  //temporary registers
  private static int am,ax;   //mantissa and exponent a
  private static int bm,bx;   //mantissa and exponent b
  private static int rm,rx;   //mantissa and exponent result
  private static boolean rneg;//true if result negative
  private static char[] buf = new char[11];

  /**
   * Create f3 representation
   * Note: the result is not normalized after creation
   *
   * @param mantissa Value -999 to +999
   * @param exponent Value -16 to +15
   * @return f3 value representing (mantissa/1000)*(10^exponent)
   */
  public static int set(int mantissa, int exponent) {
    return (mantissa<<5)|(exponent&0x1F);
  }

  /**
   * Extract mantissa from f3 value
   * Note: value f is not normalized before extracting
   *
   * @param f Value to extract mantissa from
   * @return mantissa from value f (-999 to +999)
   */
  public static int mantissa(int f) {
    return f>>5;
  }

  /**
   * Extract exponent from f3 value
   * Note: value f is not normalized before extracting
   *
   * @param f Value to extract exponent from
   * @return exponent from value f (-16 to +15)
   */
  public static int exponent(int f) {
    int e = f&0x1F;
    if (e>15) e -= 32;
    return e;
  }

  /**
   * Normalize f3 value
   * This method makes the |mantissa| as large as possible (100..999),
   * by shifting the mantissa left and decrementing the exponent, as far as
   * the exponent allows.
   * Examples: -0.010E-15 will normalize to -0.100E-16
   *            0.001E-15 will normalize to  0.010E-16
   *
   * @param f Value representing (mantissa/1000)*(10^exponent)
   * @return Normalized f3 value
   *         If the mantissa is outside the -999 to +999 interval, NaN is returned
   */
  public static int normalize(int f) {
    int fm = UnsignedIntMath.abs(mantissa(f)); // |mantissa f|
    if (fm == 0) return 0;// -+0.000E-+XX = 0.000E+00
    if (fm>999) return NaN; //NaN
    int fx = exponent(f);
    while ((fm<100) && (fx>-16)) {
      fm *= 10;
      fx--;
    }
    if (f<0) fm = -fm;
    return (fm<<5)|(fx&0x1F);
  }

  /**
   * Pre muldiv
   * Setup rneg,am,ax,bm,bx registers for mul(a,b) and div(a,b)
   *
   * @param a f3 value
   * @param b f3 value
   */
  private static void pre_muldiv(int a, int b) {
    rneg = (a^b)>>>15 != 0; //result sign (true if result negative)
    int an = normalize(a);
    int bn = normalize(b);
    am = UnsignedIntMath.abs(mantissa(an)); // |mantissa a|
    ax = exponent(an);
    bm = UnsignedIntMath.abs(mantissa(bn)); // |mantissa b|
    bx = exponent(bn);
  }

  /**
   * Post muldiv
   * Process intermediate result from mul(a,b) and div(a,b)
   *
   * @return Normalized f3 value for mul(a,b) and div(a,b)
   */
  private static int post_muldiv() {
    while ((rm<1000) && (rx>-16)) {
      rm *= 10;
      rx--;
    }
    if (rm == 0) return 0;
    if (rx<-16) return 0;
    if (rx>15) return (rneg) ? NEGATIVE_INFINITY : POSITIVE_INFINITY;
    rm /=10;
    if (rneg) rm = -rm;
    rm <<= 5;
    rx &=0x1F;
    return rm|rx;
  }

  /**
   * Multiply two f3 values
   *
   * @param a f3 value
   * @param b f3 value
   * @return Normalized f3 value for a*b
   */
  public static int mul(int a, int b) {
    pre_muldiv(a,b);
    if ((am==0) || (bm==0)) return 0;
    rm = UnsignedIntMath.umulf(am*10,UnsignedIntMath.dec2frac(bm*10)); //10*mantissa result
    rx = ax + bx;
    return post_muldiv();
  }

  /**
   * Divide two f3 values
   * Note: special values are returned when b=0
   *
   * @param a f3 value
   * @param b f3 value
   * @return Normalized f3 value for a/b
   *         if b=0 then: if a=0 then NaN is returned
   *                      if a<0 then NEGATIVE_INFINITY is returned
   *                      if a>0 then POSITIVE_INFINITY is returned
   */
  public static int div(int a, int b) {
    pre_muldiv(a,b);
    if ((am==0) && (bm!=0)) return 0;
    if (bm==0) { //division by zero
      if (am>0) return POSITIVE_INFINITY;
      return (am<0) ? NEGATIVE_INFINITY : NaN;
    }
    int ri = UnsignedIntMath.udiv(am*10,bm*10); //integer part of mantissa result
    rm = UnsignedIntMath.frac2dec(UnsignedIntMath.ufrac(am*10,bm*10)); //10*fraction result
    rx = ax - bx;
    while (ri>0) {
      rm = (rm/10)+1000*(ri%10);
      ri /= 10;
      rx++;
    }
    return post_muldiv();
  }

  /**
   * Pre addsub
   * Setup am,ax,bm,bx,rx registers for add(a,b) and sub(a,b)
   *
   * @param a f3 value
   * @param b f3 value
   */
  private static void pre_addsub(int a, int b) {
    int an = normalize(a);
    int bn = normalize(b);
    am = mantissa(an); // mantissa a
    ax = exponent(an);
    bm = mantissa(bn); // mantissa b
    bx = exponent(bn);
    while (ax>bx) {
      bm = bm/10;
      bx++;
    }
    while (ax<bx) {
      am = am/10;
      ax++;
    }
    rx = ax;
  }

  /**
   * Post addsub
   * Process intermediate result from add(a,b) and sub(a,b)
   *
   * @return Normalized f3 value for add(a,b) and sub(a,b)
   */
  private static int post_addsub() {
    if (rm == 0) return 0;
    int rn = UnsignedIntMath.abs(rm);
    if (rn>999) {
      rm = (rm+5)/10;
      rx++;
      rn = (rn+5)/10;
    }
    if (rx>15) return (rm < 0) ? NEGATIVE_INFINITY : POSITIVE_INFINITY;
    while ((rn<100) && (rx>-16)) {
      rn *= 10;
      rm *= 10;
      rx--;
    }
    rm <<= 5;
    rx &=0x1F;
    return rm|rx;
  }

  /**
   * Add two f3 values
   *
   * @param a f3 value
   * @param b f3 value
   * @return Normalized f3 value for a+b
   */
  public static int add(int a, int b) {
    pre_addsub(a,b);
    rm = am + bm;
    return post_addsub();
  }

  /**
   * Subtract two f3 values
   *
   * @param a f3 value
   * @param b f3 value
   * @return Normalized f3 value for a-b
   */
  public static int sub(int a, int b) {
    pre_addsub(a,b);
    rm = am - bm;
    return post_addsub();
  }

  /**
   * Compare two f3 values
   *
   * @param a f3 value
   * @param b f3 value
   * @return -1 if a<b, 0 if a=b, 1 if a>b
   */
  public static int compare(int a, int b) {
    int c = sub(a,b);
    if (c==0) return 0;
    if (c<0) return -1;
    return 1;
  }

  /**
   * Print f3 value to IDE message window.
   * Output format "s0.mmmEsee" where s is '+','-' or ' ' and m and e are '0' to '9'.
   * Note: value f is not normalized before printing
   *
   * @param f Value to print
   */
  public static void printf(int f) {
    int k = bprintf(buf,0,f);
    buf[k]=0; //make asciiz string
    Format.printf(buf); //write to IDE message window
  }

  /**
   * Print f3 value to char array.
   * Output format "s0.mmmEsee" where s is '+','-' or ' ' and m and e are '0' to '9'.
   * Note: value f is not normalized before printing
   *
   * @param buf Char array to write output to
   * @param offset Start index in buf
   * @param f Value to print
   * @return offset in buf after writing output
   */
  public static int bprintf(char[] buf, int offset, int f) {
    int i = mantissa(f);
    offset = Format.bprintf(buf,offset,(i<0) ? "-" : " ");
    offset = Format.bprintf(buf,offset,"0.");
    offset = Format.bprintf(buf,offset,"%03dE",UnsignedIntMath.abs(i));
    i = exponent(f);
    offset = Format.bprintf(buf,offset,(i<0) ? "-" : "+");
    offset = Format.bprintf(buf,offset,"%02d",UnsignedIntMath.abs(i));
    return offset;
  }

}
